#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// helper for 'get_s_record'
static unsigned char get_char(char *ptr)
{
	unsigned char i, ret = 0;
	for (i = 0; i < 2; i++, *ptr++)
	{
		ret <<= 4;
		if ((*ptr & 0xf0) == 0x30 && (*ptr & 0xf) < 10) ret |= (*ptr & 0xf);
		if ((*ptr & 0xd0) == 0x40 && (*ptr & 0xf) > 0 && (*ptr & 0xf) < 7) ret |= ((*ptr & 0x7) + 9);
	}

	return ret;
}

// function to import easy68k 's' records into raw data
static int get_s_record(unsigned char *dest)
{
	char line[128];

	FILE *fa = fopen("photoy2k_asm.s68", "rt");
	if (fa == NULL) {
		printf ("Failed to open photoy2k_asm.s68.\n");
		return 1;
	}

	printf ("Importing \'S\' record...\n");

	while (1)
	{
		if (fgets (line, 128, fa) == NULL) break;

		char *ptr = line;
		while (*ptr == ' ' || *ptr == '\t') ptr++;
		if (*ptr != 'S') continue; // ??
		ptr++; // past 's' (1 byte)

		if (*ptr != '1' && *ptr != '2' && *ptr != '3') continue;

		unsigned char address_size = (*ptr & 3) + 1; ptr++;
		unsigned int address = 0;
		unsigned char bytecount = get_char(ptr); ptr += 2;
		unsigned char checksum = bytecount;

		while (address_size)
		{
			address <<= 8;
			address |= get_char(ptr);
			checksum += address & 0xff;
			ptr+=2;
			address_size--;
			bytecount--;
		}

		bytecount--;

		while (bytecount)
		{
			dest[address ^ 1] = get_char(ptr);
			ptr += 2;
			checksum += dest[address ^ 1];
			address++;
			bytecount--;
		}

		unsigned char chk = get_char(ptr);
		checksum ^= 0xff;
		if (chk != checksum)
		{
		//	bprintf (0, _T("checksum failed! Expected: %2.2x Calculated: %2.2x\n"), chk, checksum ^ 0xff);			
			break;
		}
	}

	fclose (fa);
	
	return 0;
}

// remove encryption on 68k rom
static void pgm_decrypt_photoy2k(unsigned char *rom)
{
	int i;
	static const unsigned char photoy2k_tab[256] = {
		0xd9, 0x92, 0xb2, 0xbc, 0xa5, 0x88, 0xe3, 0x48, 0x7d, 0xeb, 0xc5, 0x4d, 0x31, 0xe4, 0x82, 0xbc,
		0x82, 0xcf, 0xe7, 0xf3, 0x15, 0xde, 0x8f, 0x91, 0xef, 0xc6, 0xb8, 0x81, 0x97, 0xe3, 0xdf, 0x4d,
		0x88, 0xbf, 0xe4, 0x05, 0x25, 0x73, 0x1e, 0xd0, 0xcf, 0x1e, 0xeb, 0x4d, 0x18, 0x4e, 0x6f, 0x9f,
		0x00, 0x72, 0xc3, 0x74, 0xbe, 0x02, 0x09, 0x0a, 0xb0, 0xb1, 0x8e, 0x9b, 0x08, 0xed, 0x68, 0x6d,
		0x25, 0xe8, 0x28, 0x94, 0xa6, 0x44, 0xa6, 0xfa, 0x95, 0x69, 0x72, 0xd3, 0x6d, 0xb6, 0xff, 0xf3,
		0x45, 0x4e, 0xa3, 0x60, 0xf2, 0x58, 0xe7, 0x59, 0xe4, 0x4f, 0x70, 0xd2, 0xdd, 0xc0, 0x6e, 0xf3,
		0xd7, 0xb2, 0xdc, 0x1e, 0xa8, 0x41, 0x07, 0x5d, 0x60, 0x15, 0xea, 0xcf, 0xdb, 0xc1, 0x1d, 0x4d,
		0xb7, 0x42, 0xec, 0xc4, 0xca, 0xa9, 0x40, 0x30, 0x0f, 0x3c, 0xe2, 0x81, 0xe0, 0x5c, 0x51, 0x07,
		0xb0, 0x1e, 0x4a, 0xb3, 0x64, 0x3e, 0x1c, 0x62, 0x17, 0xcd, 0xf2, 0xe4, 0x14, 0x9d, 0xa6, 0xd4,
		0x64, 0x36, 0xa5, 0xe8, 0x7e, 0x84, 0x0e, 0xb3, 0x5d, 0x79, 0x57, 0xea, 0xd7, 0xad, 0xbc, 0x9e,
		0x2d, 0x90, 0x03, 0x9e, 0x0e, 0xc6, 0x98, 0xdb, 0xe3, 0xb6, 0x9f, 0x9b, 0xf6, 0x21, 0xe6, 0x98,
		0x94, 0x77, 0xb7, 0x2b, 0xaa, 0xc9, 0xff, 0xef, 0x7a, 0xf2, 0x71, 0x4e, 0x52, 0x06, 0x85, 0x37,
		0x81, 0x8e, 0x86, 0x64, 0x39, 0x92, 0x2a, 0xca, 0xf3, 0x3e, 0x87, 0xb5, 0x0c, 0x7b, 0x42, 0x5e,
		0x04, 0xa7, 0xfb, 0xd7, 0x13, 0x7f, 0x83, 0x6a, 0x77, 0x0f, 0xa7, 0x34, 0x51, 0x88, 0x9c, 0xac,
		0x23, 0x90, 0x4d, 0x4d, 0x72, 0x4e, 0xa3, 0x26, 0x1a, 0x45, 0x61, 0x0e, 0x10, 0x24, 0x8a, 0x27,
		0x92, 0x14, 0x23, 0xae, 0x4b, 0x80, 0xae, 0x6a, 0x56, 0x01, 0xac, 0x55, 0xf7, 0x6d, 0x9b, 0x6d
	};

	unsigned short *src = (unsigned short *)rom;

	for (i = 0; i < 0x200000/2; i++) {
		unsigned short x = src[i];

		if ((i & 0x40080) != 0x00080) x ^= 0x0001;
		if ((i & 0x84008) == 0x84008) x ^= 0x0002;
		if ((i & 0x00030) == 0x00010) x ^= 0x0004;
		if ((i & 0x00242) != 0x00042) x ^= 0x0008;
		if ((i & 0x48100) == 0x48000) x ^= 0x0010;
		if ((i & 0x02004) != 0x00004) x ^= 0x0020;
		if ((i & 0x01800) != 0x00000) x ^= 0x0040;
		if ((i & 0x04820) == 0x04820) x ^= 0x0080;

		x ^= photoy2k_tab[i & 0xff] << 8;

		src[i] = x;
	}
}

// calculate all possible bitswapped sprite addresses (bottom 14 bits)
static unsigned int spritenum(unsigned short photoy2k_seqpos)
{
#define BITSWAP16(n, 														\
				  bit15, bit14, bit13, bit12, bit11, bit10, bit09, bit08,	\
				  bit07, bit06, bit05, bit04, bit03, bit02, bit01, bit00)	\
		(((((n) >> (bit15)) & 1) << 15) | 									\
		 ((((n) >> (bit14)) & 1) << 14) | 									\
		 ((((n) >> (bit13)) & 1) << 13) | 									\
		 ((((n) >> (bit12)) & 1) << 12) | 									\
		 ((((n) >> (bit11)) & 1) << 11) | 									\
		 ((((n) >> (bit10)) & 1) << 10) | 									\
		 ((((n) >> (bit09)) & 1) <<  9) | 									\
		 ((((n) >> (bit08)) & 1) <<  8) | 									\
		 ((((n) >> (bit07)) & 1) <<  7) | 									\
		 ((((n) >> (bit06)) & 1) <<  6) | 									\
		 ((((n) >> (bit05)) & 1) <<  5) | 									\
		 ((((n) >> (bit04)) & 1) <<  4) | 									\
		 ((((n) >> (bit03)) & 1) <<  3) | 									\
		 ((((n) >> (bit02)) & 1) <<  2) | 									\
		 ((((n) >> (bit01)) & 1) <<  1) | 									\
		 ((((n) >> (bit00)) & 1) <<  0))

	switch((photoy2k_seqpos >> 10) & 0xf)
	{
		case 0x0:
		case 0xa:
			return BITSWAP16(photoy2k_seqpos, 15,14,13,12,11,10, 0,8,3,1,5,9,4,2,6,7) ^ 0x124;
		case 0x1:
		case 0xb:
			return BITSWAP16(photoy2k_seqpos, 15,14,13,12,11,10, 5,1,7,4,0,8,3,6,9,2) ^ 0x088;
		case 0x2:
		case 0x8:
			return BITSWAP16(photoy2k_seqpos, 15,14,13,12,11,10, 3,5,9,7,6,4,1,8,2,0) ^ 0x011;
		case 0x3:
		case 0x9:
			return BITSWAP16(photoy2k_seqpos, 15,14,13,12,11,10, 1,8,3,6,0,4,5,2,9,7) ^ 0x154;
		case 0x4:
		case 0xe:
			return BITSWAP16(photoy2k_seqpos, 15,14,13,12,11,10, 2,1,7,4,5,8,3,6,9,0) ^ 0x0a9;
		case 0x5:
		case 0xf:
			return BITSWAP16(photoy2k_seqpos, 15,14,13,12,11,10, 9,4,6,8,2,1,7,5,3,0) ^ 0x201;
		case 0x6:
		case 0xd:
			return BITSWAP16(photoy2k_seqpos, 15,14,13,12,11,10, 4,6,0,8,9,7,3,5,1,2) ^ 0x008;
		case 0x7:
		case 0xc:
			return BITSWAP16(photoy2k_seqpos, 15,14,13,12,11,10, 8,9,3,2,0,1,6,7,5,4) ^ 0x000;
	}
	
	return 0;
}

static void apply_patches(unsigned char *src)
{
	int i,j, counter = 0;
	unsigned short *rom = (unsigned short*)src;

	// make a copy of the 68K rom into the upper 2mb
	// this is referenced by the anti-tamper routines
	memcpy (src + 0x200000, src, 0x200000);

	// find anti-tamper routines, make a copy of them elsewhere in the 68k rom, 
	// add opcodes to add / subtract address to +/- 2mb (reference the above, unmodified copy of the data)
	for (i = 0; i < 0x30000/2; i++)
	{	
		if ((rom[i] & 0xff00) == 0x6500 || (rom[i] & 0xff00) == 0x6200) // found bcs / bhi
		{
			unsigned int flags = 0;
			unsigned int offset = 0;
			unsigned int offset_bra = 0;
			unsigned int offset_bcs = 0x100000 + (i * 2);
			unsigned int offset_beq = 0;

			for (j = 0; j < 0x20; j++)
			{
				if ((rom[i-1-j] & 0xff00) == 0x6000) // found bra
				{
					flags |= 1;
					offset_bra = 0x100000 + ((i-1-j)*2) + 2;
				}
			}
			
			if (flags == 0) continue;
			
			for (j = 0; j < 0x20; j++)
			{
				if ((rom[i+1+j] & 0xff00) == 0x6600) // found bne
				{
					flags |= 2;
				}

				if ((rom[i+1+j] & 0xff00) == 0x6700) // found beq
				{
					if (flags & 2)
					{
						flags |= 4;
						offset = 0x100000 + ((i + j) * 2);
						offset_beq = i + j;
					}
					else
					{
						flags = 0;
					}
					break;
				}
			}

			if (flags == 7){
			
				int alt_cmp = ((rom[i-1] & 0xff00) != 0xff00) ? 1 : 0;
			
				int current_offset = 0xa0000 + (counter * 0x40);

				offset_bra -= 0x100000;
				offset_bcs -= 0x100000;
			
				memset (src + current_offset, 0, 0x40);
			
				int Ax = -1;
				for (j = 0; j < 8; j++)
				{
					// 3x1y                     move.w  (Ay), Dx
					if ((rom[(offset_bra/2)+j] & 0xf1f8) == 0x3010)
					{
						Ax = rom[(offset_bra/2)+j] & 7;
						break;
					}
					
					// D052                     add.w   (A2), D0
					// 9052                     sub.w   (A2), D0
					if ((rom[(offset_bra/2)+j] & 0xb1f8) == 0x9050)
					{
						Ax = rom[(offset_bra/2)+j] & 7;
						break;
					}
				}

				if (Ax == -1) { /*printf ("error! %x\n", i * 2 + 0x100000);*/ continue; }
			
				rom[(current_offset / 2) + 0] = 0xd1fc | (Ax << 9);
				rom[(current_offset / 2) + 1] = 0x20;
				rom[(current_offset / 2) + 2] = 0;
			
				current_offset += 6;
				int copy_len = (offset_bcs + 2) - offset_bra;
			
				memcpy (src + current_offset, src + offset_bra, copy_len);
				current_offset += copy_len;

				if (alt_cmp)
				{
					if (rom[(offset_bcs + 2)/2] == 0x2D6E) { // 2D6E FFF6 FFE8           move.l  (-$a,A6), (-$18,A6)
						memcpy (src + current_offset, src + (offset_bcs + 2), 6);
						current_offset += 6;
						alt_cmp++;
				//		printf ("Extra-%x, %d\n", offset_bra + copy_len, alt_cmp);
					} else {
						memcpy (src + current_offset, src + offset_bra + copy_len, 2);
						current_offset += 2;
					}
					
					offset_bcs += 2;
				}
			
				rom[(current_offset / 2) + 0] = 0x91fc | (Ax << 9);
				rom[(current_offset / 2) + 1] = 0x20;
				rom[(current_offset / 2) + 2] = 0;
				rom[(current_offset / 2) + 3] = 0x4e75;

				rom[(offset_bcs / 2) - 2] = 0x4eb9;
				current_offset = 0x1a0000 + (counter * 0x40);
				rom[(offset_bcs / 2) - 1] = current_offset >> 16;
				rom[(offset_bcs / 2) - 0] = current_offset & 0xffff;
				if (alt_cmp == 2) rom[(offset_bcs / 2) + 1] = 0x4e71;
				if (alt_cmp == 2) rom[(offset_bcs / 2) + 2] = 0x4e71;
			
			//	printf ("found: (%x) BRA: %5.5x, BCS: %5.5x, count: %d\n", flags, offset_bra, offset_bcs, counter);
				counter++;
				i = offset_beq;
			}
		}
	}

	// remove reads/writes relating to the asic
	for (i = 0; i < 0x100000/2; i++)
	{	
		if (rom[i] == 0x33c0 && rom[i+1] == 0x50 && (rom[i+2] & 0xfffd) == 0)
		{
			int which = rom[i+2] & 2; // 500000 or 500002

			if (which == 0) {
				rom[i] = 0x4e71;
				rom[i+1] = 0x4e71;
				rom[i+2] = 0x4e71;
			} else {
				rom[i] = 0x4eb9;
				rom[i+1] = 0x130000 >> 16;
				rom[i+2] = 0x130000 & 0xffff;
			}
		}

		//12666C: 33F9 0081 3566 0050 0000 move.w  $813566.l, $500000.l
		//126676: 33F9 0081 3564 0050 0002 move.w  $813564.l, $500002.l
		if (rom[i+0] == 0x33f9 && rom[i+1] == 0x081 && rom[i+2] == 0x3566 && rom[i+3] == 0x50 && rom[i+4] == 0 &&
			rom[i+5] == 0x33f9 && rom[i+6] == 0x081 && rom[i+7] == 0x3564 && rom[i+8] == 0x50 && rom[i+9] == 2)
		{
			rom[i+0] = 0x4eb9;
			rom[i+1] = 0x130000 >> 16;
			rom[i+2] = 0x130000 & 0xffff;
			rom[i+3] = rom[i+4] = rom[i+5] = rom[i+6] = rom[i+7] = rom[i+8] = rom[i+9] = 0x4e71;
		}
		
		if ((rom[i] & 0xf1ff) == 0x3039 && rom[i+1] == 0x50 && (rom[i+2] & 0xfffd) == 0)
		{
			int which = rom[i+2] & 2; // 500000 or 500002
			
			rom[i+1] = 0x81e000 >> 16;
			rom[i+2] = (0x81e000 & 0xffff) + (~which & 2); // right?
		}
		
		//126572: 4279 0050 0004           clr.w   $500004.l
		if (rom[i] == 0x4279 && rom[i+1] == 0x50 && rom[i+2] == 4)
		{
			rom[i] = rom[i+1] = rom[i+2] = 0x4e71;
		}
	}
}

// copy in data from the asic rom that is needed for the asic simulation
static void patch_arm_data(unsigned char *dest, unsigned char *src)
{
	int i;
	unsigned short *pgmsalt = (unsigned short*)(dest + 0x70000);

	// the encrypted sprite addresses are "salted" by data from the asic
	for (i = 0; i < 0x4000; i++)
	{
		unsigned char salt = src[0x1cfc + (i >> 1)] >> ((i & 1) * 4);

		pgmsalt[i] = ((salt & 0x01) << 1) | ((salt & 0x02) << 2) | ((salt & 0x04) << 6) | ((salt & 0x08) << 10);
	}

	pgmsalt += 0x4000;

	// unknown lookup table
	for (i = 0; i < 58; i++)
	{
		pgmsalt[i] = src[0x1888 + i * 4];
	}

	pgmsalt = (unsigned short*)(dest + 0xc0000);

	// make a copy of the possible bitswapped sprite addresses
	// this is much faster than calculating the bitswaps on-the-fly
	for (i = 0; i < 0x10 << 10; i++)
	{
		pgmsalt[i] = spritenum(i);
	}
}

int main()
{
	printf ("Preparing to build file...\n");

	unsigned char *PGM68KROM = (unsigned char*)malloc(0x400000);
	unsigned char *PGMARMROM = (unsigned char*)malloc(0x4000);

	FILE *fa = fopen("pgm_p0701_v105.u2", "rb");
	if (fa == NULL) {
		printf ("%s not found!\n", "pgm_p0701_v105.u2");
		free (PGM68KROM);
		free (PGMARMROM);
		return 1;
	}
	fread (PGM68KROM, 0x200000, 1, fa);
	fclose (fa);
	
	fa = fopen("igs027a_photoy2k_v100_china.asic", "rb");
	if (fa == NULL) {
		printf ("%s not found!\n", "igs027a_photoy2k_v100_china.asic");
		free (PGM68KROM);
		free (PGMARMROM);
		return 1;
	}
	fread (PGMARMROM, 0x04000, 1, fa);
	fclose (fa);

	printf ("Program data imported...\n");

	pgm_decrypt_photoy2k(PGM68KROM);
	
	printf ("Program data decrypted...\n");
	
	apply_patches(PGM68KROM);
	patch_arm_data(PGM68KROM, PGMARMROM);

	printf ("Program data patched...\n");
	
	if (get_s_record(PGM68KROM - 0x100000)) {
		free (PGM68KROM);
		free (PGMARMROM);
		return 1;
	}
	
	*((unsigned short*)(PGM68KROM + 0x3f0008)) = 0x0003; // world
	
	fa = fopen("pgm_p0701_v105.u2-deprotected", "wb");
	fwrite (PGM68KROM, 0x400000, 1, fa);
	fclose (fa);

	free(PGMARMROM);
	free(PGM68KROM);
	
	printf ("complete.\n");
	getch();
	
	return 0;
}
