*-----------------------------------------------------------
* Title         : Knights of Valour protection simulation
* Written by    : IQ_132
* Date          : 2024
* Description   : Simulate the protection routines required by this game in 68k assembly
*-----------------------------------------------------------
    ORG    $130000
START:                  ; first instruction of program

COMMAND_WORD		equ		$813564       		; 2 bytes
DATA_WORD		    equ	    $813566       		; 2 bytes

SALT_LUT			equ		$170000		; 8000
BA_LUT				equ		$178000		; 5a
AE_LUT				equ		$17805a		; 1e
BITSWAP				equ		$1c0000		; 8000

RESPONSE_VALUE 		equ		$880000

_unused_ram         equ     $81e000

_response		    equ	    _unused_ram+$0  ; 4 bytes
_key                equ     _unused_ram+$4  ; 4 bytes
_sim_value		    equ	    _unused_ram+$8	; 2 bytes

_background_select	equ		_unused_ram+$a	; 2 bytes
_text_select		equ		_unused_ram+$c	; 2 bytes

_cmd_21_value		equ		_unused_ram+$e	; 2 bytes
_cmd_22_value		equ		_unused_ram+$10	; 2 bytes
_cmd_23_value		equ		_unused_ram+$12	; 2 bytes
_seqpos				equ		_unused_ram+$14 ; 4 bytes


protection_routine
    movem.l     D0-D2/A0, -(A7)

	move.w		COMMAND_WORD, D0
    move.w      D0, D1
    lsr.w       #$8, D1
    cmp.b       #$ff, D1
    bne.s       _skip_calc_key_ff
    	move.l      #$ffffffff, _key
_skip_calc_key_ff
    move.w      _key, D1

    ; apply key to command
    eor.b       D1, D0
    andi.w      #$ff, D0

    ; apply key to sim value
	move.w		DATA_WORD, D2
	eor.w       D1, D2
	move.w      D2, _sim_value

	move.l		#RESPONSE_VALUE, _response

	cmp.b		#$99, D0
	beq			_99_command
	cmp.b		#$e0, D0
	beq			_e0_command
	cmp.b		#$ae, D0
	beq			_ae_command	
	cmp.b		#$ba, D0
	beq			_ba_command
	cmp.b		#$c0, D0
	beq			_c0_command
	cmp.b		#$c3, D0
	beq			_c3_command
	cmp.b		#$cb, D0
	beq			_cb_command
	cmp.b		#$cc, D0
	beq			_cc_command
	cmp.b		#$d0, D0
	beq			_d0_command
	cmp.b		#$dc, D0
	beq			_dc_command
	cmp.b		#$20, D0
	beq			_20_command
	cmp.b		#$21, D0
	beq			_21_command
	cmp.b		#$22, D0
	beq			_22_command
	cmp.b		#$23, D0
	beq			_23_command
	cmp.b		#$30, D0
	beq			_30_command
	cmp.b		#$32, D0
	beq			_32_command

	; fall through if not found...
end_protection_routine

    moveq.l     #$0, D0    ; update key!
    moveq.l     #$0, D1
    move.b      _key, D0
    add.w       #$1, D0
    andi.w      #$ff, D0
    cmpi.b      #$ff, D0
    bne.s       _skip_reset_key
    moveq.l     #$1, D0
_skip_reset_key
    move.b      D0, D1
    lsl.w       #$8, D0
    move.b      D1, D0
    move.w      D0, D1
    swap        D0
    move.w      D1, D0
    move.l      D0, _key

    move.l      _response, D1 ; apply key to response!
    eor.l       D0, D1
    move.l      D1, _response
    movem.l     (A7)+, D0-D2/A0
	rts



_99_command
    move.l      #$0, _key
	bra.s       end_protection_routine


_e0_command
	move.l		#$a00000, D0
	move.w		_sim_value, D0
	andi.w		#$1f, D0
	lsl.w		#$6, D0
	move.l		D0, _response
	bra	        end_protection_routine


_c0_command
	move.w		_sim_value, _text_select
	bra			end_protection_routine


_c3_command
	move.l		#$904000, D0
	moveq.l		#$0, D1
	move.w		_sim_value, D1
	lsl.l		#$8, D1
	add.l		D1, D0
	moveq.l		#$0, D1
	move.w		_text_select, D1
	lsl.l		#$2, D1
	add.l		D1, D0
	move.l		D0, _response
	bra			end_protection_routine


_cb_command
	move.w		_sim_value, _background_select
	bra			end_protection_routine


_cc_command
	move.l		#$900000, D0
	move.w		_sim_value, D0
	andi.w		#$f, D0
	lsl.w		#$8, D0
	move.w		_background_select, D1
	lsl.w		#$2, D1
	add.w		D1, D0
	move.l		D0, _response
	bra		    end_protection_routine


_d0_command
	move.l		#$a01000, D0
	moveq.l		#$0, D1
	move.w		_sim_value, D1
	lsl.w		#$5, D1
	add.w		D1, D0
	move.l		D0, _response
	bra			end_protection_routine


_dc_command
	move.l		#$a00800, D0
	moveq.l		#$0, D1
	move.w		_sim_value, D1
	lsl.w		#$6, D1
	add.w		D1, D0
	move.l		D0, _response
	bra			end_protection_routine


_20_command
	moveq.l		#$0, D0
    move.w      _cmd_21_value, D0
	move.l		D0, _response
	bra	        end_protection_routine


_21_command
	lea.l		SALT_LUT, A0
    move.w      _sim_value, _cmd_21_value
	move.w		_cmd_23_value, D1
	andi.w		#$3fff, D1
	add.w		D1, D1
	move.w		(A0,D1.w), D1
	move.w		_cmd_22_value, D0
	andi.l		#$def5, D0
	or.w		D1, D0
	move.l		D0, _response
	bra	        end_protection_routine


_22_command
    move.w      _sim_value, _cmd_22_value
	move.l		#RESPONSE_VALUE, _response
	bra	        end_protection_routine


_23_command
    move.w      _sim_value, _cmd_23_value
	move.l		#RESPONSE_VALUE, _response
	bra	        end_protection_routine


_30_command
	move.l		_seqpos, D0
	addq.l		#$1, D0
	move.l		D0, _seqpos
	bsr.s		get_seqpos
	bra			end_protection_routine


_32_command
	moveq.l		#$0, D0
	move.w		_sim_value, D0
	lsl.l		#$4, D0
	move.l		D0, _seqpos
	bsr.s		get_seqpos
	bra			end_protection_routine


get_seqpos
	lea.l		BITSWAP, A0
	; D0 is seqpos at this point
	move.w		D0, D1
	andi.w		#$c000, D0
	andi.w		#$3fff, D1
	add.w		D1, D1
	or.w		(A0,D1.w), D0
	move.l		D0, _response
	rts


_ae_command
	lea.l		AE_LUT, A0
	move.w		_sim_value, D0
	andi.l		#$f, D0
	add.b		D0, D0
	move.w		(A0,D0.w), D0
	move.l		D0, _response
	bra			end_protection_routine


_ba_command
	lea.l		BA_LUT, A0
	move.w		_sim_value, D0
	andi.l		#$3f, D0
	add.b		D0, D0
	move.w		(A0,D0.w), D0
	move.l		D0, _response
	bra			end_protection_routine

    END    START        ; last line of source



*~Font name~Courier New~
*~Font size~18~
*~Tab type~1~
*~Tab size~4~
